"""Asset of type Link."""
import logging
from typing import List

import click

from ostorlab.assets import link as link_asset
from ostorlab.cli.scan.run import run
from ostorlab.cli import console as cli_console


logger = logging.getLogger(__name__)
console = cli_console.Console()



@run.run.command()
@click.option('--url', help='List of Urls to scan.', required=True, multiple=True)
@click.option('--method', help='List of HTTP methods.', required=True, multiple=True)
@click.pass_context
def link(ctx: click.core.Context, url: List[str], method: List[str]) -> None:
    """Run scan for links."""
    runtime = ctx.obj['runtime']
    if len(url) != len(method):
        console.error('Make sure every URL has its corresponding method.')
        raise click.exceptions.Exit(2)
    assets = []
    for (u, m) in zip(url, method):
        asset = link_asset.Link(url=u, method=m)
        assets.append(asset)
    logger.debug('scanning assets %s', asset)
    runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], assets=assets)



"""Vulnz module that handles listing and describing a vulnerability."""
import click

from ostorlab.cli.rootcli import rootcli
from ostorlab.runtimes import registry


@rootcli.group()
@click.option('--runtime', type=click.Choice(['local', 'litelocal', 'cloud', 'hybrid']),
              help="""Runtime is in charge of preparing the environment to trigger a scan.\n
                    Specify which runtime to use: \n
                    local: on you local machine\n
                    litelocal: stripped down local runtime\n
                    cloud: on Ostorlab cloud, (requires login)\n
                    hybrid: running partially on Ostorlab cloud and partially on the local machine\n
                   """,
              default='local',
              required=True)
@click.pass_context
def vulnz(ctx, runtime: str = 'local') -> None:
    """You can use vulnz to list and describe vulnerabilities.\n"""
    try:
        runtime_instance = registry.select_runtime(runtime_type=runtime)
        ctx.obj['runtime'] = runtime_instance
    except registry.RuntimeNotFoundError as e:
        raise click.ClickException(f'The selected runtime {runtime} is not supported.') from e
    pass



"""Search agent API request."""

import json
from typing import Dict, Optional

from ostorlab.apis import request


class AgentSearchAPIRequest(request.APIRequest):
    """Search agent API request."""

    def __init__(self, search: str) -> None:
        """Initializer"""
        self._search = search

    @property
    def query(self) -> Optional[str]:
        """The query to fetch the agent details with an agent key.

        Returns:
            The query to fetch the agent details.
        """
        return """
            query Agents($search: String!){
                agents(search: $search) {
                    agents {
                        key
                        versions (orderBy:Version, sort:Desc, numberElements:1, page:1) {
                            versions {
                                key
                                version
                                description
                                inSelectors
                                outSelectors
                              }
              
                        }
                    }
                }
            }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the body of the API request, to fetch the specific agent.

        Returns:
            The body of the agent details request.
        """
        data = {
            'query': self.query,
            'variables': json.dumps({'search': self._search})
        }
        return data



"""Utils to handle scan risk rating."""
import enum
from typing import List

class RiskRating(enum.Enum):
    """Enumeration of the risk rating of a scan."""
    HIGH = 'High'
    MEDIUM = 'Medium'
    LOW = 'Low'
    POTENTIALLY = 'Potentially'
    HARDENING = 'Hardening'
    SECURE = 'Secure'
    IMPORTANT = 'Important'
    INFO = 'Info'

    @classmethod
    def has_value(cls, value: str) -> bool:
        return value in cls._member_map_

    @classmethod
    def values(cls) -> List[str]:
        return [key.lower() for key in cls._member_map_.keys()]


RATINGS_ORDER = {
    RiskRating.HIGH.name: 0,
    RiskRating.MEDIUM.name: 1,
    RiskRating.LOW.name: 2,
    RiskRating.POTENTIALLY.name: 3,
    RiskRating.HARDENING.name: 4,
    RiskRating.SECURE.name: 5,
    RiskRating.IMPORTANT.name: 6,
    RiskRating.INFO.name: 7
}



# pylint: disable=C0103
"""Revision reflecting the state of databases created with ostorlab<0.11.2.
Useful as a starting point for future migrations.

Revision ID: 35cd577ef0e5
Revises:
Create Date: 2022-10-26 16:00:52.792734

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '35cd577ef0e5'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('scan',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(length=255), nullable=True),
    sa.Column('asset', sa.String(length=255), nullable=True),
    sa.Column('created_time', sa.DateTime(), nullable=True),
    sa.Column('progress',
              sa.Enum('NOT_STARTED', 'IN_PROGRESS', 'STOPPED', 'DONE', 'ERROR', name='scanprogress'),
              nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('scan_status',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_time', sa.DateTime(), nullable=True),
    sa.Column('key', sa.String(length=255), nullable=True),
    sa.Column('value', sa.Text(), nullable=True),
    sa.Column('scan_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['scan_id'], ['scan.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('vulnerability',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('technical_detail', sa.Text(), nullable=True),
    sa.Column('risk_rating',
        sa.Enum('HIGH', 'MEDIUM', 'LOW', 'POTENTIALLY', 'HARDENING', 'SECURE', 'IMPORTANT', 'INFO', name='riskrating'),
        nullable=True),
    sa.Column('cvss_v3_vector', sa.String(length=1024), nullable=True),
    sa.Column('dna', sa.String(length=256), nullable=True),
    sa.Column('title', sa.String(length=256), nullable=True),
    sa.Column('short_description', sa.Text(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('recommendation', sa.Text(), nullable=True),
    sa.Column('scan_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['scan_id'], ['scan.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('vulnerability')
    op.drop_table('scan_status')
    op.drop_table('scan')
    # ### end Alembic commands ###



"""Pretty prints and logs console statements."""

from typing import Dict, List

import rich
from rich import box
from rich import status


class Console:
    """Pretty prints and logs console statements."""

    THEME = {
        'success': 'bold green',
        'error': 'red',
        'warning': 'yellow',
        'info': 'bold blue'
    }

    def __init__(self, theme: Dict[str, str] = None) -> None:
        """Initializes the console with text styling.

        Args:
            theme: The text styling. Defaults to None.
        """
        if theme is None:
            theme = self.THEME
        self._console = rich.console.Console(theme=rich.theme.Theme(theme))
        self._table = rich.table.Table

    def success(self, text: str) -> None:
        """Shows success message.

        Args:
            text: The success text to show.
        """
        self._console.print(f':heavy_check_mark: {text}', style='success')

    def error(self, text: str) -> None:
        """Shows error message.

        Args:
            text: The error text to show.
        """
        self._console.print(f':small_red_triangle: [bold]ERROR:[/] {text}', style='error')

    def warning(self, text: str) -> None:
        """Shows warning message.

        Args:
            text: The warning text to show.
        """
        self._console.print(f':small_orange_diamond: [bold]WARNING:[/] {text}', style='warning')

    def info(self, text: str) -> None:
        """Shows general information message.

        Args:
            text: The general text to show.
        """
        self._console.print(f':small_blue_diamond: {text}')

    def status(self, text: str) -> status.Status:
        """Shows loading text.

        Args:
            text: The loading text to show.

        Returns:
            The loading text.
        """
        return self._console.status(f'[info]{text}')

    def table(self, columns: Dict[str, str], data: List[Dict], title: str) -> None:
        """Constructs a table to display a list of items.

        Args:
            columns: The table columns.
            data: The list of items to display.
            title: The title of the table.
        """

        table = self._table(title=f'\n[bold]{title}', show_lines=True)

        for column in columns.keys():
            table.add_column(column)

        for item in data:
            row_values = []
            for column in columns.values():
                row_values.append(item[column])
            table.add_row(*row_values)

        table.box = box.SQUARE_DOUBLE_HEAD
        self._console.print(table)

    def print(self, data):
        self._console.print(data)



"""Android .APK asset."""
import dataclasses
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.file.android.apk')
class AndroidApk(asset.Asset):
    """Android .APK  target asset."""

    def __init__(self, content: bytes, path: Optional[str] = None):
        self.content = content
        self.path = path

    def __str__(self) -> str:
        return f'Android APK ({self.path})'

    @property
    def proto_field(self) -> str:
        return 'android_apk'



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/use/device/device.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='ostorlab/agent/message/proto/v3/use/device/device.proto',
  package='ostorlab.agent.message.proto.v3.use.device',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n7ostorlab/agent/message/proto/v3/use/device/device.proto\x12*ostorlab.agent.message.proto.v3.use.device\"3\n\rLoginPassword\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\"1\n\x06SSHKey\x12\x13\n\x0bprivate_key\x18\x01 \x01(\t\x12\x12\n\npublic_key\x18\x02 \x01(\t\"\xb8\x01\n\x0b\x43redentials\x12S\n\x0elogin_password\x18\x01 \x01(\x0b\x32\x39.ostorlab.agent.message.proto.v3.use.device.LoginPasswordH\x00\x12\x45\n\x07ssh_key\x18\x02 \x01(\x0b\x32\x32.ostorlab.agent.message.proto.v3.use.device.SSHKeyH\x00\x42\r\n\x0b\x63redentials\":\n\x0b\x44\x65viceRelay\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\x11\"*\n\x08\x44uration\x12\r\n\x05nanos\x18\x01 \x01(\x05\x12\x0f\n\x07seconds\x18\x02 \x01(\x03\"/\n\x08\x41rgument\x12\x10\n\x08\x61rg_name\x18\x01 \x01(\t\x12\x11\n\targ_value\x18\x02 \x03(\t\"\xd1\x03\n\x07Message\x12\x11\n\tdevice_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65vice_type\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65vice_name\x18\x03 \x01(\t\x12S\n\x12\x64\x65vice_credentials\x18\x04 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.Credentials\x12M\n\x0c\x64\x65vice_relay\x18\x05 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.DeviceRelay\x12Y\n\x18\x64\x65vice_relay_credentials\x18\x06 \x01(\x0b\x32\x37.ostorlab.agent.message.proto.v3.use.device.Credentials\x12\x46\n\x08\x64uration\x18\x07 \x01(\x0b\x32\x34.ostorlab.agent.message.proto.v3.use.device.Duration\x12\x42\n\x04\x61rgs\x18\x08 \x03(\x0b\x32\x34.ostorlab.agent.message.proto.v3.use.device.Argument')
)




_LOGINPASSWORD = _descriptor.Descriptor(
  name='LoginPassword',
  full_name='ostorlab.agent.message.proto.v3.use.device.LoginPassword',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='username', full_name='ostorlab.agent.message.proto.v3.use.device.LoginPassword.username', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='password', full_name='ostorlab.agent.message.proto.v3.use.device.LoginPassword.password', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=103,
  serialized_end=154,
)


_SSHKEY = _descriptor.Descriptor(
  name='SSHKey',
  full_name='ostorlab.agent.message.proto.v3.use.device.SSHKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='private_key', full_name='ostorlab.agent.message.proto.v3.use.device.SSHKey.private_key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='public_key', full_name='ostorlab.agent.message.proto.v3.use.device.SSHKey.public_key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=156,
  serialized_end=205,
)


_CREDENTIALS = _descriptor.Descriptor(
  name='Credentials',
  full_name='ostorlab.agent.message.proto.v3.use.device.Credentials',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='login_password', full_name='ostorlab.agent.message.proto.v3.use.device.Credentials.login_password', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ssh_key', full_name='ostorlab.agent.message.proto.v3.use.device.Credentials.ssh_key', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='credentials', full_name='ostorlab.agent.message.proto.v3.use.device.Credentials.credentials',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=208,
  serialized_end=392,
)


_DEVICERELAY = _descriptor.Descriptor(
  name='DeviceRelay',
  full_name='ostorlab.agent.message.proto.v3.use.device.DeviceRelay',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ostorlab.agent.message.proto.v3.use.device.DeviceRelay.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='address', full_name='ostorlab.agent.message.proto.v3.use.device.DeviceRelay.address', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='ostorlab.agent.message.proto.v3.use.device.DeviceRelay.port', index=2,
      number=3, type=17, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=394,
  serialized_end=452,
)


_DURATION = _descriptor.Descriptor(
  name='Duration',
  full_name='ostorlab.agent.message.proto.v3.use.device.Duration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='nanos', full_name='ostorlab.agent.message.proto.v3.use.device.Duration.nanos', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='seconds', full_name='ostorlab.agent.message.proto.v3.use.device.Duration.seconds', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=454,
  serialized_end=496,
)


_ARGUMENT = _descriptor.Descriptor(
  name='Argument',
  full_name='ostorlab.agent.message.proto.v3.use.device.Argument',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='arg_name', full_name='ostorlab.agent.message.proto.v3.use.device.Argument.arg_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='arg_value', full_name='ostorlab.agent.message.proto.v3.use.device.Argument.arg_value', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=498,
  serialized_end=545,
)


_MESSAGE = _descriptor.Descriptor(
  name='Message',
  full_name='ostorlab.agent.message.proto.v3.use.device.Message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='device_id', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_type', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_name', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_credentials', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_credentials', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_relay', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_relay', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_relay_credentials', full_name='ostorlab.agent.message.proto.v3.use.device.Message.device_relay_credentials', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='duration', full_name='ostorlab.agent.message.proto.v3.use.device.Message.duration', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='args', full_name='ostorlab.agent.message.proto.v3.use.device.Message.args', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=548,
  serialized_end=1013,
)

_CREDENTIALS.fields_by_name['login_password'].message_type = _LOGINPASSWORD
_CREDENTIALS.fields_by_name['ssh_key'].message_type = _SSHKEY
_CREDENTIALS.oneofs_by_name['credentials'].fields.append(
  _CREDENTIALS.fields_by_name['login_password'])
_CREDENTIALS.fields_by_name['login_password'].containing_oneof = _CREDENTIALS.oneofs_by_name['credentials']
_CREDENTIALS.oneofs_by_name['credentials'].fields.append(
  _CREDENTIALS.fields_by_name['ssh_key'])
_CREDENTIALS.fields_by_name['ssh_key'].containing_oneof = _CREDENTIALS.oneofs_by_name['credentials']
_MESSAGE.fields_by_name['device_credentials'].message_type = _CREDENTIALS
_MESSAGE.fields_by_name['device_relay'].message_type = _DEVICERELAY
_MESSAGE.fields_by_name['device_relay_credentials'].message_type = _CREDENTIALS
_MESSAGE.fields_by_name['duration'].message_type = _DURATION
_MESSAGE.fields_by_name['args'].message_type = _ARGUMENT
DESCRIPTOR.message_types_by_name['LoginPassword'] = _LOGINPASSWORD
DESCRIPTOR.message_types_by_name['SSHKey'] = _SSHKEY
DESCRIPTOR.message_types_by_name['Credentials'] = _CREDENTIALS
DESCRIPTOR.message_types_by_name['DeviceRelay'] = _DEVICERELAY
DESCRIPTOR.message_types_by_name['Duration'] = _DURATION
DESCRIPTOR.message_types_by_name['Argument'] = _ARGUMENT
DESCRIPTOR.message_types_by_name['Message'] = _MESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

LoginPassword = _reflection.GeneratedProtocolMessageType('LoginPassword', (_message.Message,), dict(
  DESCRIPTOR = _LOGINPASSWORD,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.LoginPassword)
  ))
_sym_db.RegisterMessage(LoginPassword)

SSHKey = _reflection.GeneratedProtocolMessageType('SSHKey', (_message.Message,), dict(
  DESCRIPTOR = _SSHKEY,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.SSHKey)
  ))
_sym_db.RegisterMessage(SSHKey)

Credentials = _reflection.GeneratedProtocolMessageType('Credentials', (_message.Message,), dict(
  DESCRIPTOR = _CREDENTIALS,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.Credentials)
  ))
_sym_db.RegisterMessage(Credentials)

DeviceRelay = _reflection.GeneratedProtocolMessageType('DeviceRelay', (_message.Message,), dict(
  DESCRIPTOR = _DEVICERELAY,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.DeviceRelay)
  ))
_sym_db.RegisterMessage(DeviceRelay)

Duration = _reflection.GeneratedProtocolMessageType('Duration', (_message.Message,), dict(
  DESCRIPTOR = _DURATION,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.Duration)
  ))
_sym_db.RegisterMessage(Duration)

Argument = _reflection.GeneratedProtocolMessageType('Argument', (_message.Message,), dict(
  DESCRIPTOR = _ARGUMENT,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.Argument)
  ))
_sym_db.RegisterMessage(Argument)

Message = _reflection.GeneratedProtocolMessageType('Message', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGE,
  __module__ = 'ostorlab.agent.message.proto.v3.use.device.device_pb2'
  # @@protoc_insertion_point(class_scope:ostorlab.agent.message.proto.v3.use.device.Message)
  ))
_sym_db.RegisterMessage(Message)


# @@protoc_insertion_point(module_scope)



# pylint: disable=C0103
"""Add references field to vulnerability

Revision ID: 7d12204c3f27
Revises: 746c5eb3d181
Create Date: 2022-10-25 22:47:46.208220

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7d12204c3f27'
down_revision = '746c5eb3d181'
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        'vulnerability',
        sa.Column('references', sa.Text)
    )


def downgrade() -> None:
    pass



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ostorlab/agent/message/proto/v3/asset/agent/agent.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7ostorlab/agent/message/proto/v3/asset/agent/agent.proto\x12+ostorlab.agent.message.proto.v3.asset.agent\"r\n\x07Message\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x14\n\x0cgit_location\x18\x03 \x01(\t\x12\x17\n\x0f\x64ocker_location\x18\x04 \x01(\t\x12\x1a\n\x12yaml_file_location\x18\x05 \x01(\t')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'ostorlab.agent.message.proto.v3.asset.agent.agent_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGE._serialized_start=104
  _MESSAGE._serialized_end=218
# @@protoc_insertion_point(module_scope)



"""Module offering methods to minify a dictionary,
by truncating its string & bytes values to a specific configurable size.
"""
from typing import Any, Callable, Dict, List, Union

TRUNCATE_SIZE = 256


def truncate_str(value: Union[str, bytes], ) -> Union[str, bytes]:
    """Truncate a string or bytes value.

    Args:
        s: the string or bytes value.
        truncate_size: how much to truncate.

    Returns:
        the truncated string or bytes value.
    """
    if isinstance(value, (str, bytes)):
        # The casting to string is specific to the bytes case, to prevent the json encoding from failing later.
        value = str(value)[:TRUNCATE_SIZE]
    return value


def minify_dict(value: Any, handler: Callable[[object], object]) -> Union[Dict[object, object], List[object], object]:
    """Recursive approach to minify dictionary values.

    Args:
        dic: The dictionary to minify.
        handler: Method that will be applied to all the values.

    Returns:
        the minified version of the dict.
    """
    if isinstance(value, list):
        return [minify_dict(v, handler) for v in value]
    elif isinstance(value, dict):
        for key, v in value.items():
            value[key] = minify_dict(v, handler)
        return value
    else:
        return handler(value)


def _nested_set(dic: Any, keys: List[object], value: Any) -> None:
    """Populates value in the correct place following the path."""
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value



"""Local runtime runs agents locally.

The local runtime requires Docker Swarm to run robust long-running services with a set of configured services, like
a local RabbitMQ.
"""
import logging
from typing import Dict, List
from typing import Optional
from concurrent import futures

import click
import docker
import rich
import sqlalchemy
import tenacity
from docker.models import services as docker_models_services
from rich import markdown
from rich import panel
from sqlalchemy import case

from ostorlab import exceptions
from ostorlab.assets import asset as base_asset
from ostorlab.cli import console as cli_console
from ostorlab.cli import docker_requirements_checker
from ostorlab.cli import dumpers
from ostorlab.cli import install_agent
from ostorlab.runtimes import definitions
from ostorlab.runtimes import runtime
from ostorlab.runtimes.local import agent_runtime
from ostorlab.runtimes.local import log_streamer
from ostorlab.runtimes.local.models import models
from ostorlab.runtimes.local.services import mq
from ostorlab.runtimes.local.services import redis
from ostorlab.runtimes.local.services import jaeger
from ostorlab.utils import risk_rating
from ostorlab.utils import styles
from ostorlab.utils import volumes

NETWORK_PREFIX = 'ostorlab_local_network'

logger = logging.getLogger(__name__)
console = cli_console.Console()

ASSET_INJECTION_AGENT_DEFAULT = 'agent/ostorlab/inject_asset'
TRACKER_AGENT_DEFAULT = 'agent/ostorlab/tracker'
LOCAL_PERSIST_VULNZ_AGENT_DEFAULT = 'agent/ostorlab/local_persist_vulnz'

DEFAULT_AGENTS = [
    ASSET_INJECTION_AGENT_DEFAULT,
    TRACKER_AGENT_DEFAULT,
    LOCAL_PERSIST_VULNZ_AGENT_DEFAULT
]


class UnhealthyService(exceptions.OstorlabError):
    """A service by the runtime is considered unhealthy."""


class AgentNotInstalled(exceptions.OstorlabError):
    """Agent image not installed."""


class AgentNotHealthy(exceptions.OstorlabError):
    """Agent not healthy."""


def _has_container_image(agent: definitions.AgentSettings):
    """Check if container image is available"""
    return agent.container_image is not None


def _is_service_type_run(service: docker_models_services.Service) -> bool:
    """Checks if the service should run once or should be continuously running.

    Args:
        service: Docker service.

    Returns:
        Bool indicating if the service is run-once or long-running.
    """
    return service.attrs['Spec']['TaskTemplate']['RestartPolicy']['Condition'] == 'none'


class LocalRuntime(runtime.Runtime):
    """Local runtime runs agents locally using Docker Swarm.
    Local runtime starts a Vanilla RabbitMQ service, starts all the agents listed in the `AgentRunDefinition`, checks
    their status and then inject the target asset.
    """

    def __init__(self, *args,
                 tracing: Optional[bool] = False,
                 mq_exposed_ports: Optional[Dict[int, int]] = None, **kwargs) -> None:
        super().__init__()
        del args, kwargs
        self.follow = []
        self._tracing = tracing
        self._mq_service: Optional[mq.LocalRabbitMQ] = None
        self._redis_service: Optional[redis.LocalRedis] = None
        self._jaeger_service: Optional[jaeger.LocalJaeger] = None
        self._log_streamer = log_streamer.LogStream()
        self._scan_db: Optional[models.Scan] = None
        self._mq_exposed_ports: Optional[Dict[int, int]] = mq_exposed_ports

    @property
    def name(self) -> str:
        """Local runtime instance name."""
        if self._scan_db is not None:
            return str(self._scan_db.id)
        else:
            raise ValueError('Scan not created yet')

    @property
    def network(self) -> str:
        """Local runtime network name.

        Returns:
            Local runtime network name.
        """
        return f'{NETWORK_PREFIX}_{self.name}'

    def can_run(self, agent_group_definition: definitions.AgentGroupDefinition) -> bool:
        """Checks if the runtime can run the provided agent run definition.

        Args:
            agent_group_definition: Agent and Agent group definition.

        Returns:
            Always true for the moment as the local runtime doesn't have restrictions on what it can run.
        """
        del agent_group_definition
        self._docker_checks()
        return True

    def _docker_checks(self):
        """checking the requirements (docker,swarm,arch,permissions) for ostorlab."""
        if not docker_requirements_checker.is_docker_installed():
            console.error('Docker is not installed.')
            raise click.exceptions.Exit(2)
        elif not docker_requirements_checker.is_sys_arch_supported():
            console.error('System architecture is not supported.')
            raise click.exceptions.Exit(2)
        elif not docker_requirements_checker.is_user_permitted():
            console.error('User does not have permissions to run docker.')
            raise click.exceptions.Exit(2)
        elif not docker_requirements_checker.is_docker_working():
            console.error('Error using docker.')
            raise click.exceptions.Exit(2)
        else:
            if not docker_requirements_checker.is_swarm_initialized():
                docker_requirements_checker.init_swarm()

        self._docker_client = docker.from_env()

    def scan(self, title: str, agent_group_definition: definitions.AgentGroupDefinition,
             assets: Optional[List[base_asset.Asset]]) -> None:
        """Start scan on asset using the provided agent run definition.

        The scan takes care of starting all the scan required services, ensuring they are healthy, starting all the
         agents, ensuring they are healthy and then injects the target asset.

        Args:
            title: Scan title
            agent_group_definition: Agent run definition defines the set of agents and how agents are configured.
            assets: the target asset to scan.

        Returns:
            None
        """
        try:
            console.info('Creating scan entry')
            if assets is None:
                assets_str = 'N/A'
            else:
                assets_str = f'{", ".join([str(asset) for asset in assets])}'
                # TODO(mohsinenar): we need to add support for storing multiple assets and rename this to target.
            self._scan_db = self._create_scan_db(asset=assets_str[:255], title=title)
            console.info('Creating network')
            self._create_network()
            console.info('Starting services')
            self._start_services()
            console.info('Checking services are healthy')
            self._check_services_healthy()

            console.info('Starting pre-agents')
            self._start_pre_agents()
            console.info('Checking pre-agents are healthy')
            is_healthy = self._check_agents_healthy()
            if is_healthy is False:
                raise AgentNotHealthy()

            console.info('Starting agents')
            self._start_agents(agent_group_definition)
            console.info('Checking agents are healthy')
            is_healthy = self._check_agents_healthy()
            if is_healthy is False:
                raise AgentNotHealthy()

            if assets is not None:
                self._inject_assets(assets)
            console.info('Updating scan status')
            self._update_scan_progress('IN_PROGRESS')

            console.info('Starting post-agents')
            self._start_post_agents()
            console.info('Checking post-agents are healthy')
            is_healthy = self._check_agents_healthy()
            if is_healthy is False:
                raise AgentNotHealthy()

            console.success('Scan created successfully')
        except AgentNotHealthy:
            console.error('Agent not starting')
            self.stop(self._scan_db.id)
            self._update_scan_progress('ERROR')
            self.stop(str(self._scan_db.id))
        except AgentNotInstalled as e:
            console.error(f'Agent {e} not installed')
            self.stop(str(self._scan_db.id))
        except UnhealthyService as e:
            console.error(f'Unhealthy service {e}')
            self.stop(str(self._scan_db.id))
        except agent_runtime.MissingAgentDefinitionLabel as e:
            console.error(f'Missing agent definition {e}. This is probably due to building the image directly with'
                          f' docker instead of `ostorlab agent build` command')
            self.stop(str(self._scan_db.id))

    def stop(self, scan_id: str) -> None:
        """Remove a service (scan) belonging to universe with scan_id(Universe Id).

        Args:
            scan_id: The id of the scan to stop.
        """
        try:
            int_scan_id = int(scan_id)
        except ValueError as e:
            console.error('Scan id must be an integer.')
            raise click.exceptions.Exit(2) from e

        logger.info('stopping scan id %s', scan_id)
        stopped_services = []
        stopped_network = []
        stopped_configs = []
        self._docker_checks()
        services = self._docker_client.services.list()
        for service in services:
            service_labels = service.attrs['Spec']['Labels']
            logger.info('comparing %s and %s', service_labels.get('ostorlab.universe'), scan_id)
            if service_labels.get('ostorlab.universe') == scan_id:
                stopped_services.append(service)
                service.remove()

        networks = self._docker_client.networks.list()
        for network in networks:
            network_labels = network.attrs['Labels']
            if network_labels is not None and network_labels.get('ostorlab.universe') == scan_id:
                logger.info('removing network %s', network_labels)
                stopped_network.append(network)
                network.remove()

        configs = self._docker_client.configs.list()
        for config in configs:
            config_labels = config.attrs['Spec']['Labels']
            if config_labels.get('ostorlab.universe') == scan_id:
                logger.info('removing config %s', config_labels)
                stopped_configs.append(config)
                config.remove()

        if stopped_services or stopped_network or stopped_configs:
            console.success('All scan components stopped.')

        with models.Database() as session:
            scan = session.query(models.Scan).get(int_scan_id)
            if scan:
                scan.progress = 'STOPPED'
                session.commit()
                console.success('Scan stopped successfully.')
            else:
                console.info(f'Scan {scan_id} was not found.')

    def _create_scan_db(self, title: str, asset: str):
        """Persist the scan in the database"""
        return models.Scan.create(title=title, asset=asset)

    def _update_scan_progress(self, progress: str):
        """Update scan status to in progress"""
        with models.Database() as session:
            scan = session.query(models.Scan).get(self._scan_db.id)
            scan.progress = progress
            session.commit()

    def _create_network(self):
        """Creates a docker swarm network where all services and agents can communicate."""
        if any(network.name == self.network for network in self._docker_client.networks.list()):
            logger.warning('network already exists.')
        else:
            logger.info('creating private network %s', self.network)
            return self._docker_client.networks.create(
                name=self.network,
                driver='overlay',
                attachable=True,
                labels={'ostorlab.universe': self.name},
                check_duplicate=True
            )

    def _start_services(self):
        """Start all the local runtime services."""
        self._start_mq_service()
        self._start_redis_service()
        if self._tracing is True:
            self._start_jaeger_service()

    def _start_mq_service(self):
        """Start a local rabbitmq service."""
        self._mq_service = mq.LocalRabbitMQ(name=self.name, network=self.network, exposed_ports=self._mq_exposed_ports)
        self._mq_service.start()
        if 'mq' in self.follow:
            self._log_streamer.stream(self._mq_service.service)

    def _start_redis_service(self):
        """Start a local Redis service."""
        self._redis_service = redis.LocalRedis(name=self.name, network=self.network)
        self._redis_service.start()
        if 'redis' in self.follow:
            self._log_streamer.stream(self._redis_service.service)

    def _start_jaeger_service(self):
        """Start a local Jaeger service."""
        self._jaeger_service = jaeger.LocalJaeger(name=self.name, network=self.network)
        self._jaeger_service.start()
        click.launch(self._jaeger_service.management_interface_ui)
        if 'jaeger' in self.follow:
            self._log_streamer.stream(self._jaeger_service.service)

    def _check_services_healthy(self):
        """Check if the rabbitMQ service is running and healthy."""
        if self._mq_service is None or self._mq_service.is_healthy is False:
            raise UnhealthyService('MQ service is unhealthy.')
        if self._redis_service is None or self._redis_service.is_healthy is False:
            raise UnhealthyService('Redis service is unhealthy.')
        if self._tracing is True and (self._jaeger_service is None or self._jaeger_service.is_healthy is False):
            raise UnhealthyService('Jaeger service is unhealthy.')

    def _check_agents_healthy(self):
        """Checks if an agent is healthy."""
        return self._are_agents_ready()

    def _start_agents(self, agent_group_definition: definitions.AgentGroupDefinition):
        """Starts all the agents as list in the agent run definition."""
        with futures.ThreadPoolExecutor() as executor:
            future_to_agent = {executor.submit(self._start_agent, agent, extra_configs=[]): agent for agent in
                               agent_group_definition.agents}
            for future in futures.as_completed(future_to_agent):
                future.result()

    def _start_pre_agents(self):
        """Starting pre-agents that must exist before other agents. This applies to all persistence
        agents that can start sending data at the start of the agent."""
        self._start_persist_vulnz_agent()

    def _start_post_agents(self):
        """Starting post-agents that must exist after other agents. This applies to the tracker
        that needs to monitor other agents."""
        self._start_tracker_agent()

    def _start_agent(self, agent: definitions.AgentSettings,
                     extra_configs: Optional[List[docker.types.ConfigReference]] = None,
                     extra_mounts: Optional[List[docker.types.Mount]] = None
                     ) -> None:
        """Start agent based on provided definition.

        Args:
            agent: An agent definition containing all the settings of how agent should run and what arguments to pass.
        """
        logger.info('starting agent %s with %s', agent.key, agent.args)

        if _has_container_image(agent) is False:
            raise AgentNotInstalled(agent.key)

        runtime_agent = agent_runtime.AgentRuntime(
            agent, self.name, self._docker_client, self._mq_service, self._redis_service, self._jaeger_service)
        agent_service = runtime_agent.create_agent_service(self.network, extra_configs, extra_mounts)
        if agent.key in self.follow:
            self._log_streamer.stream(agent_service)

        if agent.replicas > 1:
            self._scale_service(agent_service, agent.replicas)

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=12),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.outcome,
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _is_service_healthy(self, service: docker_models_services.Service, replicas=None) -> bool:
        """Checks if a docker service is healthy by checking all tasks status."""
        logger.debug('checking Spec service %s', service.name)
        try:
            if not replicas:
                replicas = service.attrs['Spec']['Mode']['Replicated']['Replicas']
            return replicas == len([task for task in service.tasks() if task['Status']['State'] == 'running'])
        except docker.errors.NotFound:
            return False

    def _list_agent_services(self):
        """List the services of type agents. All agent service must start with agent_."""
        services = self._docker_client.services.list(filters={'label': f'ostorlab.universe={self.name}'})
        for service in services:
            if service.name.startswith('agent_'):
                yield service

    def _start_tracker_agent(self):
        """Start the tracker agent to handle the scan lifecycle."""
        tracker_agent_settings = definitions.AgentSettings(key=TRACKER_AGENT_DEFAULT)
        self._start_agent(agent=tracker_agent_settings, extra_configs=[])

    def _start_persist_vulnz_agent(self):
        """Start the local persistence agent to dump vulnerabilities in the local config."""
        persist_vulnz_agent_settings = definitions.AgentSettings(
            key=LOCAL_PERSIST_VULNZ_AGENT_DEFAULT,
            mounts=[])
        self._start_agent(agent=persist_vulnz_agent_settings, extra_configs=[])

    def _inject_assets(self, assets: List[base_asset.Asset]):
        """Injects the scan target assets."""
        contents = {}
        for i, asset in enumerate(assets):
            console.info(f'Injecting asset: {asset}')
            contents[f'asset.binproto_{i}'] = asset.to_proto()
            contents[f'selector.txt_{i}'] = asset.selector.encode()

        volumes.create_volume(f'asset_{self.name}', contents)

        inject_asset_agent_settings = definitions.AgentSettings(key=ASSET_INJECTION_AGENT_DEFAULT,
                                                                restart_policy='none')
        self._start_agent(agent=inject_asset_agent_settings,
                          extra_mounts=[
                              docker.types.Mount(
                                  target='/asset', source=f'asset_{self.name}', type='volume'
                              )])

    def _scale_service(self, service: docker_models_services.Service, replicas: int) -> None:
        """Calling scale directly on the service causes an API error. This is a workaround that simulates refreshing
         the service object, then calling the scale API."""
        for s in self._docker_client.services.list():
            if s.name == service.name:
                s.scale(replicas)

    def list(self, page: int = 1, number_elements: int = 10) -> List[runtime.Scan]:
        """Lists scans managed by runtime.

        Args:
            page: Page number for list pagination (default 1).
            number_elements: count of elements to show in the listed page (default 10).

        Returns:
            List of scan objects.
        """
        if page is not None:
            console.warning('Local runtime ignores scan list pagination')

        scans = {}
        with models.Database() as session:
            for scan in session.query(models.Scan):
                scans[scan.id] = runtime.Scan(
                    id=scan.id,
                    asset=scan.asset,
                    created_time=scan.created_time,
                    progress=scan.progress.value,
                )

        universe_ids = set()
        client = docker.from_env()
        services = client.services.list()

        for s in services:
            try:
                service_labels = s.attrs['Spec']['Labels']
                ostorlab_universe_id = service_labels.get('ostorlab.universe')
                if 'ostorlab.universe' in service_labels.keys() and ostorlab_universe_id not in universe_ids:
                    universe_ids.add(ostorlab_universe_id)
                    if ostorlab_universe_id.isnumeric() and int(ostorlab_universe_id) not in scans:
                        console.warning(f'Scan {ostorlab_universe_id} has not traced in DB.')
            except KeyError:
                logger.warning('The label ostorlab.universe do not exist.')

        return list(scans.values())

    @tenacity.retry(stop=tenacity.stop_after_attempt(20),
                    wait=tenacity.wait_exponential(multiplier=1, max=20),
                    # return last value and don't raise RetryError exception.
                    retry_error_callback=lambda lv: lv.outcome,
                    retry=tenacity.retry_if_result(lambda v: v is False))
    def _are_agents_ready(self, fail_fast=True) -> bool:
        """Checks that all agents are ready and healthy while taking into account the run type of agent
         (once vs long-running)."""
        logger.info('listing services ...')
        agent_services = list(self._list_agent_services())
        for service in agent_services:
            logger.info('checking %s ...', service.name)
            if not _is_service_type_run(service):
                if self._is_service_healthy(service):
                    logger.info('agent service %s is healthy', service.name)
                else:
                    logger.error('agent service %s is not healthy', service.name)
                    if fail_fast:
                        return False
        return True

    def install(self) -> None:
        """Installs the default agents.

        Returns:
            None
        """
        for agent_key in DEFAULT_AGENTS:
            install_agent.install(agent_key=agent_key)

    def list_vulnz(self, scan_id: int):
        try:
            with models.Database() as session:
                vulnerabilities = session.query(models.Vulnerability).filter_by(scan_id=scan_id). \
                    order_by(models.Vulnerability.title).all()
            vulnz_list = []
            for vulnerability in vulnerabilities:
                vulnerability_location = vulnerability.location or ''
                vulnz_list.append({
                    'id': str(vulnerability.id),
                    'risk_rating': styles.style_risk(vulnerability.risk_rating.value.upper()),
                    'cvss_v3_vector': vulnerability.cvss_v3_vector,
                    'title': vulnerability.title,
                    'short_description': markdown.Markdown(vulnerability.short_description),
                    'location': markdown.Markdown(vulnerability_location),
                })

            columns = {
                'Id': 'id',
                'Title': 'title',
                'Vulnerable target': 'location',
                'Risk rating': 'risk_rating',
                'CVSS V3 Vector': 'cvss_v3_vector',
                'Short Description': 'short_description',
            }
            title = f'Scan {scan_id}: Found {len(vulnz_list)} vulnerabilities.'
            console.table(columns=columns, data=vulnz_list, title=title)
            if len(vulnz_list) == 0:
                console.info('0 vulnerabilities were found.')
            else:
                console.success('Vulnerabilities listed successfully.')

        except sqlalchemy.exc.OperationalError:
            console.error(f'scan with id {scan_id} does not exist.')

    def _print_vulnerability(self, vulnerability):
        """Print vulnerability details"""
        if vulnerability is None:
            return

        vulnerability_location = vulnerability.location or ''
        vulnz_list = [
            {'id': str(vulnerability.id),
             'risk_rating': styles.style_risk(vulnerability.risk_rating.value.upper()),
             'cvss_v3_vector': vulnerability.cvss_v3_vector,
             'title': vulnerability.title,
             'short_description': markdown.Markdown(vulnerability.short_description),
             'location': markdown.Markdown(vulnerability_location),
             }
        ]
        columns = {
            'Id': 'id',
            'Title': 'title',
            'Vulnerable target': 'location',
            'Risk rating': 'risk_rating',
            'CVSS V3 Vector': 'cvss_v3_vector',
            'Short Description': 'short_description',
        }
        title = f'Describing vulnerability {vulnerability.id}'
        console.table(columns=columns, data=vulnz_list, title=title)
        rich.print(panel.Panel(markdown.Markdown(vulnerability.description), title='Description'))
        rich.print(panel.Panel(markdown.Markdown(vulnerability.recommendation), title='Recommendation'))
        if vulnerability.references is not None:
            rich.print(panel.Panel(markdown.Markdown(vulnerability.references), title='References'))
        rich.print(panel.Panel(markdown.Markdown(vulnerability.technical_detail), title='Technical details'))

    def describe_vuln(self, scan_id: int, vuln_id: int):
        try:
            with models.Database() as session:
                vulnerabilities = []
                if vuln_id is not None:
                    vulnerability = session.query(models.Vulnerability).get(vuln_id)
                    vulnerabilities.append(vulnerability)
                elif scan_id is not None:
                    vulnerabilities = session.query(models.Vulnerability).filter_by(scan_id=scan_id). \
                        order_by(models.Vulnerability.title).all()
                for v in vulnerabilities:
                    self._print_vulnerability(v)
                console.success('Vulnerabilities listed successfully.')
        except sqlalchemy.exc.OperationalError:
            console.error('Vulnerability / scan not Found.')

    def dump_vulnz(self, scan_id: int, dumper: dumpers.VulnzDumper):
        """Dump found vulnerabilities of a scan in a specific format."""
        vulnerabilities = []
        with models.Database() as session:
            severity_sort_logic = case(value=models.Vulnerability.risk_rating,
                                       whens=risk_rating.RATINGS_ORDER).label('severity')
            vulnerabilities = session.query(models.Vulnerability).filter_by(scan_id=scan_id). \
                order_by(severity_sort_logic).all()
        vulnz_list = []
        for vulnerability in vulnerabilities:
            vuln = {
                'id': vulnerability.id,
                'risk_rating': vulnerability.risk_rating.value,
                'cvss_v3_vector': vulnerability.cvss_v3_vector,
                'title': vulnerability.title,
                'short_description': vulnerability.short_description,
                'description': vulnerability.description,
                'recommendation': vulnerability.recommendation,
                'references': vulnerability.references,
                'technical_detail': vulnerability.technical_detail,
                'location': vulnerability.location
            }
            vulnz_list.append(vuln)
        dumper.dump(vulnz_list)
        console.success(f'{len(vulnerabilities)} Vulnerabilities saved to  : {dumper.output_path}')



"""Lists the remote scans."""

from typing import Dict, Optional
import json

from ostorlab.apis import request


class ScansListAPIRequest(request.APIRequest):
    """Lists the remote scans."""

    def __init__(self, page: int, elements: int):
        self._page = page
        self._elements = elements

    @property
    def query(self) -> Optional[str]:
        """Defines the query to list the scans.

        Returns:
            The query to list the scans.
        """
        return """
         query Scans($page: Int, $numberElements: Int) {
            scans(page: $page, numberElements: $numberElements) {
                pageInfo {
                    hasNext
                    hasPrevious
                    count
                    numPages
                }
                scans {
                    assetType
                    riskRating
                    version
                    packageName
                    id
                    progress
                    createdTime
                    scanProfile
                }
            }
         }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query to list the scans.

        Returns:
              The query to list the scans.
        """
        data = {
            'query': self.query,
            'variables': json.dumps({'page': self._page, 'numberElements': self._elements})
            }
        return data



"""Lists the remote vulnz."""

import json
from typing import Dict, Optional

from ostorlab.apis import request


class VulnzListAPIRequest(request.APIRequest):
    """Lists the remote vulnz of a scan."""

    def __init__(self, scan_id: int, number_elements: int, page: int):
        self._scan_id = scan_id
        self._number_elements = number_elements
        self._page = page

    @property
    def query(self) -> Optional[str]:
        """Defines the query to list the vulnz.

        Returns:
            The query to list the vulnz.
        """
        return """
        query Scan($scanId: Int!,$page:Int,$numberElements:Int) {
            scan(scanId: $scanId) {
                vulnerabilities (page:$page,numberElements:$numberElements){
                    pageInfo{
                        hasNext
                        numPages
                    }
                    vulnerabilities{
                        id
                        technicalDetail
                        vulnerabilityLocation
                        detail{
                          title
                          shortDescription
                          description
                          recommendation
                          cvssV3Vector
                          riskRating
                        }
                    }
                }
            }
        }
        """

    @property
    def data(self) -> Optional[Dict]:
        """Sets the query to list the vulnz.

        Returns:
              The query to list the vulnz.
        """
        data = {
            'query': self.query,
            'variables': json.dumps(
                {'scanId': self._scan_id, 'page': self._page, 'numberElements': self._number_elements})
        }
        return data



# pylint: disable=C0103
"""add vulnerability location

Revision ID: 746c5eb3d181
Revises:
Create Date: 2022-10-24 16:35:29.881269

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '746c5eb3d181'
down_revision = '35cd577ef0e5'
branch_labels = None
depends_on = '35cd577ef0e5'


def upgrade() -> None:
    op.add_column(
        'vulnerability',
        sa.Column('location', sa.Text)
    )


def downgrade() -> None:
    pass
